SystemOrganization addCategory: #Marea!
SystemOrganization addCategory: #'Marea-Swapping'!
SystemOrganization addCategory: #'Marea-Serializers'!

TestCase subclass: #GraphTraverserTest
	instanceVariableNames: 'serializedObjects stack'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Marea-Swapping'!

!GraphTraverserTest methodsFor: 'as yet unclassified' stamp: 'MarianoMartinezPeck 9/13/2011 10:49'!
testCycle
	| a set |
	
	a := ClassWith2Var new.
	a var1: Date today.
	a var2: a.
	
	set := GraphTraverser traverse: a.
	self assert: set size equals: 9.! !

!GraphTraverserTest methodsFor: 'as yet unclassified' stamp: 'MarianoMartinezPeck 9/13/2011 10:43'!
testNoCycle
	| a b set |
	
	a := ClassWith2Var new.
	b := ClassWith1Var new. 
	b var1: 'nose'.
	a var1: Date today.
	a var2: b.
	
	set := GraphTraverser traverse: a.
	self assert: set size equals: 15.! !

TestCase subclass: #ObjectGraphSwapperTest
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Marea-Swapping'!

!ObjectGraphSwapperTest methodsFor: 'as yet unclassified' stamp: 'MarianoMartinezPeck 9/9/2011 14:45'!
setUp
	ObjectGraphSwapper resetGlobalTable.
	ObjectGraphSwapper resetGraphID.
	! !

!ObjectGraphSwapperTest methodsFor: 'as yet unclassified' stamp: 'MarianoMartinezPeck 9/13/2011 11:45'!
tearDown
	ObjectGraphSwapperTest compileAll ! !

!ObjectGraphSwapperTest methodsFor: 'as yet unclassified' stamp: 'MarianoMartinezPeck 9/14/2011 20:12'!
testAmountOfProxiesAndBecome
	"We test that we creat the right amount of proxies, that is, that after a GC runs, we should only have proxies for the root and the facade objects. In addition, we test we correctly did the become. "
	| a b proxiesCount |

	3 timesRepeat: [Smalltalk garbageCollect].
	proxiesCount := MareaProxyMock instanceCount.

	a := ClassWith2Var new.
	b := ClassWith1Var new. 
	"Do not put a string or something like that because otherwise it is stored in the literals of the method and hence it is considered as a shared object"
	b var1: 6.
	a var1: (Point x: 4 y: 2).
	a var2: b.

	ObjectGraphSwapper new swapOutGraph: a.

	3 timesRepeat: [Smalltalk garbageCollect].
	"One proxy for the root A and one proxy for the shared object B."
	self assert: MareaProxyMock instanceCount equals: proxiesCount + 2.
	self assert: a class = MareaProxyMock. 
	self assert: b class = MareaProxyMock. 
! !

!ObjectGraphSwapperTest methodsFor: 'as yet unclassified' stamp: 'MarianoMartinezPeck 9/14/2011 20:11'!
testFacadeObjectNotAnymoreNoIntersection
	"This tests that a graph is loaded back when the ROOT receives a message and when there is NO intersection probem, that is, there is no swapping of external graphs that refer to our facade objects. In addition, it tests that a facade object to which we have created a proxy, is garbage collected because nobody else point to it. "
	| a b proxiesCount |
	
	3 timesRepeat: [Smalltalk garbageCollect].
	proxiesCount := MareaProxyMock instanceCount.
	
	a := ClassWith2Var new.
	b := ClassWith1Var new. 
	"Do not put a string or something like that because otherwise it is stored in the literals of the method and hence it is considered as a shared object"
	b var1: 6.
	a var1: (Point x: 4 y: 2).
	a var2: b.

	ObjectGraphSwapper new swapOutGraph: a.
	3 timesRepeat: [Smalltalk garbageCollect].
	
	"Now lets say that nobody else point to the facade object B"	
	self assert: MareaProxyMock instanceCount equals: proxiesCount + 2.
	b := nil.
	3 timesRepeat: [Smalltalk garbageCollect].
	self assert: MareaProxyMock instanceCount equals: proxiesCount + 1.
	
	ObjectGraphSwapper new swapInGraphFor: a.
	3 timesRepeat: [Smalltalk garbageCollect].
	self assert: MareaProxyMock instanceCount equals: proxiesCount.
	self assert: a var2 var1 equals: 6.
	self assert: a var1 equals: 4@2.
	! !

!ObjectGraphSwapperTest methodsFor: 'as yet unclassified' stamp: 'MarianoMartinezPeck 9/14/2011 19:46'!
testRootObjectNotAnymoreNoIntersection
	"This tests that a graph is loaded back when a facade object receives a message and when there is NO intersection probem, that is, there is no swapping of external graphs that refer to our facade objects. In addition, it tests that the root object to which we have created a proxy, is garbage collected because nobody else point to it. "
	| a b proxiesCount |
	
	3 timesRepeat: [Smalltalk garbageCollect].
	proxiesCount := MareaProxyMock instanceCount.
	
	a := ClassWith2Var new.
	b := ClassWith1Var new. 
	b var1: 'nose'.
	a var1: Date today.
	a var2: b.

	ObjectGraphSwapper new swapOutGraph: a.
	3 timesRepeat: [Smalltalk garbageCollect].
	
	"Now lets say that nobody else point to the root object A"	
	self assert: MareaProxyMock instanceCount equals: proxiesCount + 2.
	a := nil.
	3 timesRepeat: [Smalltalk garbageCollect].
	self assert: MareaProxyMock instanceCount equals: proxiesCount + 1.
	
	ObjectGraphSwapper new swapInGraphFor: b.
	3 timesRepeat: [Smalltalk garbageCollect].
	self assert: MareaProxyMock instanceCount equals: proxiesCount.
	self assert: b var1 equals: 'nose'.
! !

!ObjectGraphSwapperTest methodsFor: 'as yet unclassified' stamp: 'MarianoMartinezPeck 9/14/2011 19:46'!
testUninstallFromRootNoIntersection
	"This tests that a graph is loaded back when the ROOT receives a message and when there is NO intersection probem, that is, there is no swapping of external graphs that refer to our facade objects."
	| a b proxiesCount |
	
	3 timesRepeat: [Smalltalk garbageCollect].
	proxiesCount := MareaProxyMock instanceCount.
	
	a := ClassWith2Var new.
	b := ClassWith1Var new. 
	b var1: 'nose'.
	a var1: Date today.
	a var2: b.

	ObjectGraphSwapper new swapOutGraph: a.
	3 timesRepeat: [Smalltalk garbageCollect].
	ObjectGraphSwapper new swapInGraphFor: a.
	3 timesRepeat: [Smalltalk garbageCollect].
	self assert: MareaProxyMock instanceCount equals: proxiesCount.
	self assert: b == a var2.
	self assert: b var1 equals: 'nose'.! !

!ObjectGraphSwapperTest methodsFor: 'as yet unclassified' stamp: 'MarianoMartinezPeck 9/12/2011 11:02'!
testUninstallFromSharedObjectNoIntersection
	"This tests that a graph is loaded back when a facade object receives a message and when there is NO intersection probem, that is, there is no swapping of external graphs that refer to our facade objects."
	| a b proxiesCount |
	
	3 timesRepeat: [Smalltalk garbageCollect].
	proxiesCount := MareaProxyMock instanceCount.
	
	a := ClassWith2Var new.
	b := ClassWith1Var new. 
	b var1: 'nose'.
	a var1: Date today.
	a var2: b.

	ObjectGraphSwapper new swapOutGraph: a.
	3 timesRepeat: [Smalltalk garbageCollect].
	ObjectGraphSwapper new swapInGraphFor: b.
	3 timesRepeat: [Smalltalk garbageCollect].
	self assert: MareaProxyMock instanceCount equals: proxiesCount.
	self assert: b == a var2.
	self assert: b var1 equals: 'nose'.! !

FLMapper subclass: #FLMareaProxyMapper
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Marea-Serializers'!

!FLMareaProxyMapper methodsFor: 'accessing' stamp: 'MarianoMartinezPeck 7/12/2011 11:35'!
accepts: anObject

	^ anObject mareaIsProxy! !

!FLMareaProxyMapper methodsFor: 'visiting' stamp: 'MarianoMartinezPeck 7/12/2011 11:35'!
classAndTraitCollection
	"Returns the proper collection for collecting classes and traits. Superclasses should be before of the subclasses."

	self flag: #todo. "not nice"
	^ SortedCollection sortBlock: [ :a :b | a printCompactHierarchy <= b printCompactHierarchy ]! !

!FLMareaProxyMapper methodsFor: 'evaluating' stamp: 'MarianoMartinezPeck 7/12/2011 11:35'!
internalObjects: aCollection! !

!FLMareaProxyMapper methodsFor: 'evaluating' stamp: 'MarianoMartinezPeck 7/18/2011 10:48'!
mapAndTrace: anObject
	self visitObject: anObject.
	! !

!FLMareaProxyMapper methodsFor: 'visiting' stamp: 'MarianoMartinezPeck 7/12/2011 11:35'!
visitClassOrTrait: aClassOrTrait
	"Visits a Class or a Trait that should be fully serialized."

	analyzer
		mapAndTrace: aClassOrTrait
		to: aClassOrTrait fuelCluster
		into: analyzer clustersWithInternalMetalevelObjects
		blockForNewCollection: [ self classAndTraitCollection ]! !

!FLMareaProxyMapper methodsFor: 'visiting' stamp: 'MarianoMartinezPeck 7/12/2011 11:35'!
visitMetaclass: aMetaclass

	analyzer  
		mapAndTrace: aMetaclass
		to: aMetaclass fuelCluster 
		into: analyzer clustersWithBaselevelObjects ! !

!FLMareaProxyMapper methodsFor: 'visiting' stamp: 'MarianoMartinezPeck 7/18/2011 11:54'!
visitObject: anObject
	"Visits an object in the most generic way"

	analyzer
		mapAndTrace: anObject
		to: anObject fuelCluster
		into: analyzer clustersWithBaselevelObjects! !

!FLMareaProxyMapper methodsFor: 'visiting' stamp: 'MarianoMartinezPeck 7/12/2011 11:35'!
visitWellKnownObject: anObject
	"Visits these singleton objects. They has a special treatment because they are so common. And also nil is special because it cannot be added to a Set like the other objects."

	analyzer 
		mapAndTrace: anObject 
		to: FLWellKnownObjectsCluster instance 
		into: analyzer clustersWithBaselevelObjects 
		blockForNewCollection: [ OrderedCollection new ]! !

FLFixedObjectCluster subclass: #FLProxyCluster
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Marea-Serializers'!

!FLProxyCluster classSide methodsFor: 'accessing' stamp: 'MarianoMartinezPeck 7/18/2011 12:04'!
id

	^102! !

!FLProxyCluster methodsFor: 'serialize/materialize' stamp: 'MarianoMartinezPeck 7/18/2011 11:51'!
postMaterializeReferencesOfAll: aCollection with: aMaterializer
! !

FLAnalyzer subclass: #FLMareaAnalyzer
	instanceVariableNames: 'proxiesDict'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Marea-Swapping'!

!FLMareaAnalyzer methodsFor: 'as yet unclassified' stamp: 'MarianoMartinezPeck 9/12/2011 10:31'!
createProxyFor: anObject
	^ MareaProxyMock new.! !

!FLMareaAnalyzer methodsFor: 'as yet unclassified' stamp: 'MarianoMartinezPeck 9/9/2011 10:33'!
initialize
	super initialize.
	proxiesDict := IdentityDictionary new.! !

!FLMareaAnalyzer methodsFor: 'as yet unclassified' stamp: 'MarianoMartinezPeck 9/12/2011 10:25'!
mapAndTrace: anObject
	"Map an object to its cluster. Trace its references."
	super mapAndTrace: anObject.
	"self halt."
	"proxiesDict at: anObject put: (self createProxyFor: anObject)."
	"Transcript show: 'anObject'; cr."! !

!FLMareaAnalyzer methodsFor: 'as yet unclassified' stamp: 'MarianoMartinezPeck 9/9/2011 10:47'!
proxiesDict
	^ proxiesDict! !

FLMaterializer subclass: #FLMaterializerForSwapping
	instanceVariableNames: 'weakSerializedObjects rootProxy'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Marea-Serializers'!

!FLMaterializerForSwapping classSide methodsFor: 'as yet unclassified' stamp: 'MarianoMartinezPeck 8/6/2011 11:49'!
on: aReadStream serializedObjects: serializedObjects rootProxy: rootProxy

	^self new
		stream: aReadStream;
		weakSerializedObjects: serializedObjects ;
		rootProxy: rootProxy;
		yourself
! !

!FLMaterializerForSwapping methodsFor: 'as yet unclassified' stamp: 'MarianoMartinezPeck 8/2/2011 11:05'!
addAllMaterialized: aCollection
	aCollection do: [:each | self addMaterialized: each].
! !

!FLMaterializerForSwapping methodsFor: 'as yet unclassified' stamp: 'MarianoMartinezPeck 8/6/2011 11:58'!
addMaterialized: anObject
 	| serializedInstanceOrNil |

  	serializedInstanceOrNil := weakSerializedObjects at: (materializedInstances size + 1).
	
	(serializedInstanceOrNil == rootProxy) ifTrue: [
		"the root was becomed a proxy so I cannot send any message to it because otherwise it will try to materialize it, generating a loop. 
		In addition, if this is the proxy, I ALWAYS have to take the object serialized instead of the proxy."
		materializedInstances addLast: anObject.
		^ self].
	
	serializedInstanceOrNil mareaIsProxy ifTrue: [
		materializedInstances addLast: serializedInstanceOrNil .
		^ self.].   
	
	materializedInstances addLast: (  serializedInstanceOrNil isNil
            ifTrue: [anObject]         
		 " The GC collected that object, hence we kept the one we have serialized."
            ifFalse: [serializedInstanceOrNil]
	"The GC did not collected the object, so it was probably a shared obejct. So, we kept the instance already existing in the image")! !

!FLMaterializerForSwapping methodsFor: 'as yet unclassified' stamp: 'MarianoMartinezPeck 8/2/2011 10:14'!
initialize 
	super initialize.
	weakSerializedObjects := WeakOrderedCollection new.! !

!FLMaterializerForSwapping methodsFor: 'accessing' stamp: 'MarianoMartinezPeck 8/6/2011 11:53'!
rootProxy
	^ rootProxy! !

!FLMaterializerForSwapping methodsFor: 'accessing' stamp: 'MarianoMartinezPeck 8/6/2011 11:53'!
rootProxy: anObject
	rootProxy := anObject! !

!FLMaterializerForSwapping methodsFor: 'accessing' stamp: 'MarianoMartinezPeck 8/2/2011 10:06'!
weakSerializedObjects
	^ weakSerializedObjects! !

!FLMaterializerForSwapping methodsFor: 'accessing' stamp: 'MarianoMartinezPeck 8/2/2011 10:06'!
weakSerializedObjects: anObject
	weakSerializedObjects := anObject! !

Object subclass: #AbstractMareaSerializer
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Marea-Serializers'!

!AbstractMareaSerializer methodsFor: 'API' stamp: 'MarianoMartinezPeck 4/1/2011 16:11'!
serializeAndWriteObject: anObject to: aFileName
	self shouldBeImplemented ! !

!AbstractMareaSerializer methodsFor: 'API' stamp: 'MarianoMartinezPeck 4/1/2011 16:11'!
unserializeObjectFrom: aFileName
	self shouldBeImplemented ! !

AbstractMareaSerializer subclass: #FuelCachedMareaSerializer
	instanceVariableNames: 'cachedAnalyzer'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Marea-Serializers'!

!FuelCachedMareaSerializer classSide methodsFor: 'as yet unclassified' stamp: 'MarianoMartinezPeck 7/6/2011 11:48'!
newDefault
	^ self new cachedAnalyzer: FLAnalyzer newDefault! !

!FuelCachedMareaSerializer methodsFor: 'accessing' stamp: 'MarianoMartinezPeck 7/6/2011 11:48'!
cachedAnalyzer
	^ cachedAnalyzer! !

!FuelCachedMareaSerializer methodsFor: 'accessing' stamp: 'MarianoMartinezPeck 7/6/2011 11:48'!
cachedAnalyzer: anObject
	cachedAnalyzer := anObject! !

!FuelCachedMareaSerializer methodsFor: 'reset' stamp: 'MarianoMartinezPeck 7/6/2011 11:51'!
resetedCachedAnalyzer
	^ self cachedAnalyzer reset
	! !

!FuelCachedMareaSerializer methodsFor: 'reset' stamp: 'MarianoMartinezPeck 7/6/2011 14:48'!
resetedCachedAnalyzerWithInternalObjects: aCollection
	^ self cachedAnalyzer resetWithInternalObjects: aCollection
	! !

!FuelCachedMareaSerializer methodsFor: 'read - write' stamp: 'MarianoMartinezPeck 7/12/2011 13:31'!
serializeAndWriteClass: aClass to: aFileName
| serializer aFileStream |
      aFileStream := (StandardFileStream fileNamed: aFileName) binary.
      serializer := FLSerializer on: aFileStream.
	serializer 
		serialize: aClass 
		analyzer: (self resetedCachedAnalyzerWithInternalObjects: (Array with: aClass)).
      aFileStream close.
      self mareaLog: 'The class ', aClass name, ' was written on file'.! !

!FuelCachedMareaSerializer methodsFor: 'read - write' stamp: 'MarianoMartinezPeck 7/12/2011 13:36'!
serializeAndWriteObject: anObject to: aFileName
	| serializer aFileStream | 
	aFileStream := (StandardFileStream fileNamed: aFileName) binary.
	serializer := FLSerializer on: aFileStream.
	serializer serialize: anObject analyzer: self resetedCachedAnalyzer.
	aFileStream close.
	self mareaLog: 'The object ', anObject mareaBasicName, ' was written on file'.
! !

!FuelCachedMareaSerializer methodsFor: 'read - write' stamp: 'MarianoMartinezPeck 7/6/2011 11:41'!
unserializeObjectFrom: aFileName
	| materializer aFileStream |
	aFileStream := (StandardFileStream fileNamed: aFileName) binary.
	materializer := FLMaterializer on: aFileStream.
	materializer materialize.
	aFileStream close.
	^ materializer root
! !

AbstractMareaSerializer subclass: #FuelMareaSerializer
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Marea-Serializers'!

!FuelMareaSerializer methodsFor: 'read - write' stamp: 'MarianoMartinezPeck 8/3/2011 17:01'!
serializeAndWriteClass: aClass to: aFileName
| serializer aFileStream |
      aFileStream := (StandardFileStream fileNamed: aFileName) binary.
      serializer := FLSerializerForSwapping on: aFileStream.
      serializer
               serialize: aClass
               analyzer: (FLAnalyzer newDefaultInternalObjects: (Array with: aClass)).
      aFileStream close.
      self mareaLog: 'The class ', aClass name, ' was written on file'.
	^ serializer weakSerializedObjects ! !

!FuelMareaSerializer methodsFor: 'read - write' stamp: 'MarianoMartinezPeck 8/2/2011 10:47'!
serializeAndWriteObject: anObject to: aFileName
	| serializer aFileStream | 
	aFileStream := (StandardFileStream fileNamed: aFileName) binary.
	serializer := FLSerializerForSwapping on: aFileStream.
	serializer serialize: anObject.
	aFileStream close.
	self mareaLog: 'The object ', anObject mareaBasicName, ' was written on file'.
	^ serializer weakSerializedObjects 
! !

!FuelMareaSerializer methodsFor: 'read - write' stamp: 'MarianoMartinezPeck 4/1/2011 16:11'!
unserializeObjectFrom: aFileName
	| materializer aFileStream |
	aFileStream := (StandardFileStream fileNamed: aFileName) binary.
	materializer := FLMaterializer on: aFileStream.
	materializer materialize.
	aFileStream close.
	^ materializer root
! !

!FuelMareaSerializer methodsFor: 'read - write' stamp: 'MarianoMartinezPeck 8/6/2011 11:50'!
unserializeObjectFrom: aFileName  serializedObjects:  serializedObjects rootProxy: rootProxy
	
	| materializer aFileStream |
	aFileStream := (StandardFileStream fileNamed: aFileName) binary.
	materializer := FLMaterializerForSwapping on: aFileStream serializedObjects: serializedObjects rootProxy: rootProxy.
	materializer materialize.
	aFileStream close.
	^ materializer root
! !

Object subclass: #CompiledMethodSwapper
	instanceVariableNames: ''
	classVariableNames: 'Logger ProxyClass'
	poolDictionaries: ''
	category: 'Marea-Swapping'!

!CompiledMethodSwapper classSide methodsFor: 'stuff to exclude' stamp: 'MarianoMartinezPeck 5/3/2011 17:44'!
classesNotToSwap

	| tooDangerousClasses coreClasses protoObjectSubclasses classesSpecialArray veryUsedClasses allClasses debuggingClasses nilSubclasses |

	"This are classes that are known as too dangerous."
	tooDangerousClasses := OrderedCollection new.
	(ClassBuilder new tooDangerousClasses) 
		do: [:each | Smalltalk globals at: each 
			ifPresent: [:present | tooDangerousClasses add: (Smalltalk at: each) ] 
			ifAbsent: []].
	
	"These classes because they are too core and used by the code to swapout itseld"
	coreClasses :=  {Array. Object. ProtoObject. Class. Message. MethodDictionary. CompiledMethod}. 

	"Subclasses of ProtoObject because they do not implement some methods that Object does and that are used by the swapping stuff."
	protoObjectSubclasses := ProtoObject subclasses.
	
	"These classes cannot be swapped because they are used by the VM"
	classesSpecialArray := Smalltalk specialObjectsArray select: [ :each | each isBehavior ].
	
	"These classes cannot be swapped because they install the segment (they are used) even before the segment can be written into a file"
	veryUsedClasses := { Cursor. CursorWithMask. EncodedCharSet.}.
	
	"Classes that are usually needed to debug and they help us to analyze problems"
	debuggingClasses := { DEVToolSet. }.
	
	"nil subclasses cannot be swapped out because since they don't subclass ProtoObject they may not even implement doesNotUnderstand:.
	For example, if we swap a CompiledMethod of PseudoContext, and then we try to load it back, the VM will crash with the error Recursive not understood error encountered. '"
	nilSubclasses := nil subclasses.
	
	allClasses := coreClasses, protoObjectSubclasses, classesSpecialArray, veryUsedClasses, tooDangerousClasses, debuggingClasses, nilSubclasses.
	
	^ allClasses asOrderedCollection ! !

!CompiledMethodSwapper classSide methodsFor: 'stuff to include' stamp: 'MarianoMartinezPeck 5/13/2011 10:28'!
compiledMethodsToSwap
	|  methodsToSwap classesNotToSwap packagesNotToSwap fixedMethodsNotToSwap |
	
	fixedMethodsNotToSwap := self fixedMethodsNotToSwap.
	
	UnusedObjectDiscoverer current startDiscovery. 

	self discoverMinimalUsedCompiledMethods.
	self methodsUsedByMareaTesting.
	
	packagesNotToSwap := self packagesNotToSwap.
	
	classesNotToSwap := self classesNotToSwap.
	packagesNotToSwap do: [:aPackage | classesNotToSwap addAll: aPackage classes ].
	packagesNotToSwap do: [:aPackage | fixedMethodsNotToSwap addAll: aPackage extensionMethods ].
	
	fixedMethodsNotToSwap do: [:aMethod | 	UnusedObjectDiscoverer current primitiveSetUsedBit: aMethod boolean: true].
	
	classesNotToSwap do: [:aClass | aClass markAllInstanceAndClassMethodsAsUsed].
	
	UnusedObjectDiscoverer current stopDiscovery.
	
	methodsToSwap := CompiledMethod unusedInstances.
	methodsToSwap := methodsToSwap reject: [:each | each isInstalled not ].
	
	Transcript show: 'Methods to swap: ',  methodsToSwap size asString; cr.
	Transcript show: 'Methods NOT to swap: ',  CompiledMethod usedInstances size asString; cr.
	
	^ methodsToSwap


! !

!CompiledMethodSwapper classSide methodsFor: 'stuff to exclude' stamp: 'MarianoMartinezPeck 11/27/2010 17:05'!
discoverMinimalUsedCompiledMethods
"These the minimal CompiledMethods needed to install and install a CompiledMethodProxy"
	| aProxy |
	
	SystemNavigation default browseClass: Object.
	
	aProxy := self installOn: ClassWith1Var selector: #foo.
	(ClassWith1Var methodDict at: #foo) literals.
	
	aProxy := self installOn: ClassWith1Var selector: #foo.
	ClassWith1Var new foo.! !

!CompiledMethodSwapper classSide methodsFor: 'examples' stamp: 'MarianoMartinezPeck 7/6/2011 11:39'!
exampleSwapOutAllUnusedMethods 
	"self exampleSwapOutAllUnusedMethods"
	| methodsToSwap timeToInstall timeToUninstall counter |
	"CompiledMethodProxyHandler mareaSerializer: FuelMareaSerializer. "
	CompiledMethodProxySwapperHandler serializer: FuelMareaSerializer new. 
	self proxyClass: NumberAsFilenameCompiledMethodProxySwapper.
	self proxyClass runningTest: false.
	self proxyClass mareaResetNumber.
	Logger := nil.
	"MareaLogger changeToFileLogger."
	MareaLogger disableLogging.
	3 timesRepeat: [Smalltalk garbageCollect].
	self log: 'Using proxy class: ', self proxyClass name.
	self log: 'Proxies installed instances before start: ', self proxyClass mareaAllInstalledProxiesCount asString.
	self log: 'CompiledMethod instances before start: ', CompiledMethod instanceCount asString.
	self log: 'Memory used before proxies: ', (Smalltalk vmParameterAt: 3) asString.
	"self printSpaceTally.
"	methodsToSwap := self compiledMethodsToSwap.
	self proxyClass: NumberAsFilenameCompiledMethodProxySwapper.
	self proxyClass runningTest: false.
	
	
	CompiledMethodProxySwapperHandler serializer: FuelMareaSerializer. 
	self log: 'CompiledMethods to swap: ',  methodsToSwap size asString.
	
	counter := 0.
	"[timeToInstall := MessageTally time: ["
	methodsToSwap do: [:method| 
		"(counter \\ 5000) isZero ifTrue: [
			self log: ''."
			"self log: SmalltalkImage current vmStatisticsReportString."
			"self log: ''."
		"]. 
		counter := counter + 1."
		"self log: 'Class: ', method methodClass name asString, ' selector: ', method selector asString. "
		self installOn: method methodClass selector: method selector.].
	"]."

	self log: 'Time to install proxies: ', timeToInstall asString, ' miliseconds'.
	methodsToSwap := nil.
	MCMethodDefinition shutDown. 
	3 timesRepeat: [Smalltalk garbageCollect].
	self log: 'Proxies intalled: ', self proxyClass mareaAllInstalledProxiesCount asString.
	self log: 'CompiledMethod instances after installing: ', CompiledMethod instanceCount asString.
	self log: 'Memory used using proxies: ', (Smalltalk vmParameterAt: 3) asString.

	[
"	timeToUninstall := MessageTally time: [
		counter := 0.
		self proxyClass mareaAllInstalledProxies do: [:instance| 
			(counter \\ 5000) isZero ifTrue: [
				self log: ''.
				self log: SmalltalkImage current vmStatisticsReportString.
				self log: ''.
				self printSpaceTally.
			]. 
		counter := counter + 1.
		instance mareaUninstallIfNeeded
		].		
	]."

	"self proxyClass mareaUninstallIfNeededAllInstances."
	self log: 'Time to uninstall proxies: ', timeToUninstall asString, ' miliseconds'.
	MCMethodDefinition shutDown. 
	3 timesRepeat: [Smalltalk garbageCollect].
	self log: 'Proxies installed instances after uninstalling: ', self proxyClass mareaAllInstalledProxiesCount asString.
	self log: 'Memory used after uninstalling proxies: ', (Smalltalk vmParameterAt: 3) asString.
	self log: 'CompiledMethod instances after uninstalling: ', CompiledMethod instanceCount asString.
	"self printSpaceTally."
	] ensure: [
		self log: 'Ensure....'.
		"self proxyClass mareaUninstallIfNeededAllInstances."
"		(FileDirectory default fileNamesMatching: '*.swap*') 
			do: [ :aFile | FileDirectory default deleteFileNamed: aFile ]."
	]! !

!CompiledMethodSwapper classSide methodsFor: 'examples' stamp: 'MarianoMartinezPeck 7/6/2011 11:39'!
exampleSwapOutAllUnusedMethodsAndClasses
	"self nose"
	| methodsToSwap timeToInstall timeToUninstall counter |
	"CompiledMethodProxyHandler mareaSerializer: FuelMareaSerializer. "
	CompiledMethodProxySwapperHandler serializer: FuelMareaSerializer new. 
	self proxyClass: NumberAsFilenameCompiledMethodProxySwapper.
	self proxyClass runningTest: false.
	self proxyClass mareaResetNumber.
	Logger := nil.
	MareaLogger disableLogging.
	3 timesRepeat: [Smalltalk garbageCollect].
	self log: 'Using proxy class: ', self proxyClass name.
	self log: 'Methods Proxies installed instances before start: ', self proxyClass mareaAllInstalledProxiesCount asString.
	self log: 'Class Proxies installed instances before start: ', ClassProxySwapper mareaAllInstalledProxiesCount asString.
	self log: 'CompiledMethod instances before start: ', CompiledMethod instanceCount asString.
	self log: 'Classes instances before start: ', Smalltalk allClassesAndTraits size asString.
	self log: 'Memory used before proxies: ', (Smalltalk vmParameterAt: 3) asString.
	"self printSpaceTally.
"	methodsToSwap := self compiledMethodsToSwap.
	
	CompiledMethodProxySwapperHandler serializer: FuelMareaSerializer. 
	self log: 'CompiledMethods to swap: ',  methodsToSwap size asString.
	
	counter := 0.
	[timeToInstall := MessageTally time: [
	methodsToSwap do: [:method| 
		(counter \\ 5000) isZero ifTrue: [
			self log: ''.
			"self log: SmalltalkImage current vmStatisticsReportString."
			self log: ''.
		]. 
		counter := counter + 1.
		self installOn: method methodClass selector: method selector.].
	].

	Smalltalk allClassesDo: [:aClass | aClass hasAllMethodsAsProxies ifTrue: [ClassProxySwapper installOn: aClass ]. ].

	self log: 'Time to install proxies: ', timeToInstall asString, ' miliseconds'.
	methodsToSwap := nil.
	MCMethodDefinition shutDown. 
	3 timesRepeat: [Smalltalk garbageCollect].
	self log: 'Methods Proxies intalled: ', self proxyClass mareaAllInstalledProxiesCount asString.
	self log: 'CompiledMethod instances after installing: ', CompiledMethod instanceCount asString.
	self log: 'Class Proxies intalled: ',  ClassProxySwapper  mareaAllInstalledProxiesCount asString.
	self log: 'Class instances after installing: ', Smalltalk allClassesAndTraits size asString.
	
	self log: 'Memory used using proxies: ', (Smalltalk vmParameterAt: 3) asString.

"	timeToUninstall := MessageTally time: [
		counter := 0.
		self proxyClass mareaAllInstalledProxies do: [:instance| 
			(counter \\ 5000) isZero ifTrue: [
				self log: ''.
				self log: SmalltalkImage current vmStatisticsReportString.
				self log: ''.
				self printSpaceTally.
			]. 
		counter := counter + 1.
		instance mareaUninstallIfNeeded
		].		
	]."
	self log: 'Time to uninstall proxies: ', timeToUninstall asString, ' miliseconds'.
	MCMethodDefinition shutDown. 
	3 timesRepeat: [Smalltalk garbageCollect].
	self log: 'Method Proxies installed instances after uninstalling: ', self proxyClass mareaAllInstalledProxiesCount asString.
	self log: 'Class Proxies installed instances after uninstalling: ', ClassProxySwapper  mareaAllInstalledProxiesCount asString.
	self log: 'Memory used after uninstalling proxies: ', (Smalltalk vmParameterAt: 3) asString.
	self log: 'CompiledMethod instances after uninstalling: ', CompiledMethod instanceCount asString.
	self log: 'Class instances after uninstalling: ', Smalltalk allClassesAndTraits size asString.
	"self printSpaceTally."
	] ensure: [
		self log: 'Ensure....'.
		MareaLogger changeToFileLogger.
		"self proxyClass mareaUninstallIfNeededAllInstances."
"		(FileDirectory default fileNamesMatching: '*.swap*') 
			do: [ :aFile | FileDirectory default deleteFileNamed: aFile ]."
	]! !

!CompiledMethodSwapper classSide methodsFor: 'examples' stamp: 'MarianoMartinezPeck 7/6/2011 15:43'!
exampleSwapOutUnusedTestClasses
	"self exampleSwapOutUnusedTestClasses"
	| classesToSwap timeToInstall timeToUninstall |
	ClassProxySwapperHandler serializer: FuelCachedMareaSerializer newDefault. 
	self proxyClass: ClassProxySwapper.
	self proxyClass runningTest: false.
	"self proxyClass mareaResetNumber."
	Logger := nil.
	MareaLogger disableLogging.
	3 timesRepeat: [Smalltalk garbageCollect].
	Symbol compactSymbolTable.
	self log: 'Using proxy class: ', self proxyClass name.
	self log: 'Proxies installed instances before start: ', self proxyClass mareaAllInstalledProxiesCount asString.
	self log: 'Metaclass instances before start: ', Metaclass instanceCount asString.
	self log: 'Memory used before proxies: ', (Smalltalk vmParameterAt: 3) asString.
	classesToSwap := OrderedCollection new.
(PackageInfo allPackages select: [:each| each packageName includesSubString: 'Test'] ) do: [:aPackage |
		aPackage classes do: [:aClass | (UnusedObjectDiscoverer current primitiveGetUsedBit: aClass) 
			ifFalse: [classesToSwap add: aClass ]. 
			].
		].
	self log: 'Metaclasss to swap: ',  classesToSwap size asString.
	[
	timeToInstall := MessageTally time: [
	classesToSwap do: [:class| self installOn: class ].
	].
	self log: 'Time to install proxies: ', timeToInstall asString, ' miliseconds'.
	classesToSwap := nil.
	MCMethodDefinition shutDown. 
	3 timesRepeat: [Smalltalk garbageCollect].
	self log: 'Proxies intalled: ', self proxyClass mareaAllInstalledProxiesCount asString.
	self log: 'Metaclass instances after installing: ', Metaclass instanceCount asString.
	Symbol compactSymbolTable.
	self log: 'Memory used using proxies after compacting symbol table: ', (Smalltalk vmParameterAt: 3) asString.
	self log: 'Memory used using proxies: ', (Smalltalk vmParameterAt: 3) asString.
	timeToUninstall := MessageTally time: [self proxyClass mareaUninstallIfNeededAllInstances].
	self log: 'Time to uninstall proxies: ', timeToUninstall asString, ' miliseconds'.
	MCMethodDefinition shutDown. 
	3 timesRepeat: [Smalltalk garbageCollect].
	self log: 'Proxies installed instances after uninstalling: ', self proxyClass mareaAllInstalledProxiesCount asString.
	self log: 'Memory used after uninstalling proxies: ', (Smalltalk vmParameterAt: 3) asString.
	self log: 'Metaclass instances after uninstalling: ', Metaclass instanceCount asString.
	] ensure: [
		self proxyClass mareaUninstallIfNeededAllInstances.
		(FileDirectory default fileNamesMatching: '*.swap*') 
			do: [ :aFile | FileDirectory default deleteFileNamed: aFile ].
	]! !

!CompiledMethodSwapper classSide methodsFor: 'examples' stamp: 'MarianoMartinezPeck 7/6/2011 15:02'!
exampleSwapOutUnusedTestMethods
	"self exampleSwapOutUnusedTestMethods"
	| classesToSwap timeToInstall timeToUninstall |
	CompiledMethodProxySwapperHandler serializer: FuelCachedMareaSerializer newDefault. 
	self proxyClass: ClassProxySwapper.
	self proxyClass runningTest: false.
	self proxyClass mareaResetNumber.
	Logger := nil.
	MareaLogger disableLogging.
	3 timesRepeat: [Smalltalk garbageCollect].
	Symbol compactSymbolTable.
	self log: 'Using proxy class: ', self proxyClass name.
	self log: 'Proxies installed instances before start: ', self proxyClass mareaAllInstalledProxiesCount asString.
	self log: 'CompiledMethod instances before start: ', CompiledMethod instanceCount asString.
	self log: 'Memory used before proxies: ', (Smalltalk vmParameterAt: 3) asString.
	classesToSwap := OrderedCollection new.
(PackageInfo allPackages select: [:each| each packageName includesSubString: 'Test'] ) do: [:aPackage |
		aPackage classes do: [:aClass |
			classesToSwap addAll: aClass unusedMethods
			].
		].
	self log: 'CompiledMethods to swap: ',  classesToSwap size asString.
	[
	timeToInstall := MessageTally time: [
	classesToSwap do: [:method| self installOn: method methodClass selector: method selector.].
	].
	self log: 'Time to install proxies: ', timeToInstall asString, ' miliseconds'.
	classesToSwap := nil.
	MCMethodDefinition shutDown. 
	3 timesRepeat: [Smalltalk garbageCollect].
	self log: 'Proxies intalled: ', self proxyClass mareaAllInstalledProxiesCount asString.
	self log: 'CompiledMethod instances after installing: ', CompiledMethod instanceCount asString.
	Symbol compactSymbolTable.
	self log: 'Memory used using proxies after compacting symbol table: ', (Smalltalk vmParameterAt: 3) asString.
	self log: 'Memory used using proxies: ', (Smalltalk vmParameterAt: 3) asString.
	timeToUninstall := MessageTally time: [self proxyClass mareaUninstallIfNeededAllInstances].
	self log: 'Time to uninstall proxies: ', timeToUninstall asString, ' miliseconds'.
	MCMethodDefinition shutDown. 
	3 timesRepeat: [Smalltalk garbageCollect].
	self log: 'Proxies installed instances after uninstalling: ', self proxyClass mareaAllInstalledProxiesCount asString.
	self log: 'Memory used after uninstalling proxies: ', (Smalltalk vmParameterAt: 3) asString.
	self log: 'CompiledMethod instances after uninstalling: ', CompiledMethod instanceCount asString.
	] ensure: [
		self proxyClass mareaUninstallIfNeededAllInstances.
		(FileDirectory default fileNamesMatching: '*.swap*') 
			do: [ :aFile | FileDirectory default deleteFileNamed: aFile ].
	]! !

!CompiledMethodSwapper classSide methodsFor: 'stuff to exclude' stamp: 'MarianoMartinezPeck 5/3/2011 18:00'!
fixedMethodsNotToSwap
	^ self selectorsNotToSwap
		inject: OrderedCollection new
		into: [:allMethods :aSelector | 
			allMethods addAll: 
				(CompiledMethod allInstances select: [:aMethod | aMethod selector = aSelector] ); yourself 
		]! !

!CompiledMethodSwapper classSide methodsFor: 'install' stamp: 'MarianoMartinezPeck 7/6/2011 15:39'!
installOn: aClass 
	^ self proxyClass installOn: aClass! !

!CompiledMethodSwapper classSide methodsFor: 'install' stamp: 'MarianoMartinezPeck 11/21/2010 16:27'!
installOn: aClass selector:  aSelector.
	^ self proxyClass installOn: aClass selector: aSelector.! !

!CompiledMethodSwapper classSide methodsFor: 'lists' stamp: 'MarianoMartinezPeck 11/21/2010 18:32'!
listOfCorePackages
	^ self listOfCorePackagesNames collect: [:each |  PackageInfo named: each ]! !

!CompiledMethodSwapper classSide methodsFor: 'lists' stamp: 'MarianoMartinezPeck 11/21/2010 18:32'!
listOfCorePackagesNames
	"To recreate this list, you can take a PharoCore image and print:
	PackageInfo allPackages collect: [:each| each packageName] 
"

	^  #('MultilingualTests' 'Morphic' 'ToolBuilder-Kernel' 'PackageInfo' 'Polymorph-ToolBuilder' 'Network-Kernel' 'Network-Protocols' 'Compiler' 'System-Finalization' 'Settings-Graphics' 'Collections' 'Network-RemoteDirectory' 'Traits' 'Network-URI' 'Settings-Display' 'Collections-Abstract' 'Settings-Compiler' 'MonticelloMocks' 'ScriptLoader11' 'System-Digital Signatures' 'System-Localization' 'MenuRegistration' 'Polymorph-TaskbarIcons' 'Monticello' 'ToolBuilder-Morphic' 'MCDirtyPackage' 'Collections-Weak' 'Settings-Polymorph' 'Tests' 'HelpSystem-Core-Utilities' 'SUnit' 'Collections-Text' 'Compression' 'Morphic-MorphTreeWidget' 'HelpSystem-Core' 'ToolsTest' 'Gofer-Core' 'Gofer-Tests' 'Collections-Stack' 'FreeType' 'MonticelloConfigurations' 'Polymorph-Geometry' 'Collections-Support' 'SUnitGUI' 'HelpSystem-Core-Model' 'Tests-VM' 'System-Change Notification' 'System' 'Tests-ObjectsAsMethods' 'HostMenus' 'Network-MIME' 'Network-RFC822' 'VB-Regex' 'TrueType' 'DeprecatedPreferences' 'System-Clipboard' 'System-Settings' 'MorphicTests' 'System-Serial Port' 'NetworkTests' 'Collections-Unordered' 'System-Changes' 'System-Object Storage' 'Collections-Traits' 'System-Platforms' 'Tests-PrimCallController' 'Morphic-TrueType' 'Multilingual' 'Graphics' 'Collections-Strings' 'KernelTests' 'System-FileRegistry' 'Tests-Bugs' 'ToolBuilder-SUnit' 'Tests-Release' 'Morphic-Balloon' 'Settings-Network' 'System-Applications' 'Polymorph-Tools-Diff' 'Settings-Tools' 'ScriptLoader' 'ST80' 'FreeTypeSubPixelAntiAliasing' 'Settings-Kernel' 'Settings-System' 'CompilerTests' 'System-Tools' 'System-Hashing' 'System-Download' 'Collections-Streams' 'Tests-SystemChangeNotification' 'GraphicsTests' 'Polymorph-Widgets' 'Kernel' 'Tools' 'Settings-FreeType' 'FreeTypeTests' 'Settings-Monticello' 'HelpSystem-Tests' 'Polymorph-EventEnhancements' 'System-Support' 'MonticelloGUI' 'System-FilePackage' 'System-Object Events' 'Announcements' 'System-Text' 'Exceptions' 'Network-UUID' 'System-Pools' 'FixUnderscores' 'HelpSystem-Core-UI' 'GraphicsResources' 'Files' 'Gofer' 'Network-MailSending' 'Collections-Arrayed' 'CollectionsTests' 'Balloon' 'Network-Url' 'Collections-Sequenceable')! !

!CompiledMethodSwapper classSide methodsFor: 'lists' stamp: 'MarianoMartinezPeck 11/21/2010 18:32'!
listOfNotDangerousCorePackages
	^ self listOfNotDangerousCorePackagesNames collect: [:each |  PackageInfo named: each ]! !

!CompiledMethodSwapper classSide methodsFor: 'lists' stamp: 'MarianoMartinezPeck 11/22/2010 21:40'!
listOfNotDangerousCorePackagesNames
	"To recreate this list, you can take a PharoCore image and print:
	PackageInfo allPackages collect: [:each| each packageName] 
"

	^ #('Morphic' 'ToolBuilder-Kernel' 'PackageInfo' 'Polymorph-ToolBuilder' 'Network-Kernel' 'Network-Protocols' 'Compiler' 'System-Finalization' 'Settings-Graphics' 'Collections' 'Network-RemoteDirectory' 'Traits' 'Network-URI' 'Settings-Display' 'Collections-Abstract' 'Settings-Compiler' 'MonticelloMocks' 'ScriptLoader11' 'System-Digital Signatures' 'System-Localization' 'MenuRegistration' 'Polymorph-TaskbarIcons' 'Monticello' 'ToolBuilder-Morphic' 'MCDirtyPackage' 'Collections-Weak' 'Settings-Polymorph' 'HelpSystem-Core-Utilities' 'SUnit' 'Collections-Text' 'Compression' 'Morphic-MorphTreeWidget' 'HelpSystem-Core' 'Gofer-Core' 'Collections-Stack' 'FreeType' 'MonticelloConfigurations' 'Polymorph-Geometry' 'Collections-Support' 'SUnitGUI' 'HelpSystem-Core-Model' 'System-Change Notification' 'System' 'HostMenus' 'Network-MIME' 'Network-RFC822' 'VB-Regex' 'TrueType' 'DeprecatedPreferences' 'System-Clipboard' 'System-Settings' 'System-Serial Port' 'Collections-Unordered' 'System-Changes' 'System-Object Storage' 'Collections-Traits' 'System-Platforms' 'Morphic-TrueType' 'Multilingual' 'Graphics' 'Collections-Strings' 'System-FileRegistry' 'ToolBuilder-SUnit' 'Morphic-Balloon' 'Settings-Network' 'System-Applications' 'Polymorph-Tools-Diff' 'Settings-Tools' 'ScriptLoader' 'ST80' 'FreeTypeSubPixelAntiAliasing' 'Settings-Kernel' 'Settings-System' 'System-Tools' 'System-Hashing' 'System-Download' 'Collections-Streams' 'Polymorph-Widgets' 'Kernel' 'Tools' 'Settings-FreeType' 'Settings-Monticello' 'Polymorph-EventEnhancements' 'System-Support' 'MonticelloGUI' 'System-FilePackage' 'System-Object Events' 'Announcements' 'System-Text' 'Exceptions' 'Network-UUID' 'System-Pools' 'FixUnderscores' 'HelpSystem-Core-UI' 'GraphicsResources' 'Files' 'Gofer' 'Network-MailSending' 'Collections-Arrayed' 'Balloon' 'Network-Url' 'Collections-Sequenceable')! !

!CompiledMethodSwapper classSide methodsFor: 'logging' stamp: 'MarianoMartinezPeck 11/30/2010 13:27'!
log: aString
	self logger log: aString ! !

!CompiledMethodSwapper classSide methodsFor: 'logging' stamp: 'MarianoMartinezPeck 11/30/2010 13:32'!
logger	
	Logger isNil ifTrue: [
		Logger := FixesFileNameLoggerStrategy named: 'Swapper.txt'.
		Logger startLogging].
	^ Logger! !

!CompiledMethodSwapper classSide methodsFor: 'stuff to exclude' stamp: 'MarianoMartinezPeck 11/30/2010 18:09'!
methodsUsedByMareaTesting
	| suite |
	suite := TestSuite named: 'zaraza'. 
	AbstractCompiledMethodProxyTest withAllSubclasses do: [:each |
		TestCase addTestsFor: each name toSuite: suite.].
	suite tests: (suite tests reject: [:each | each testSelector = #testCompiledMethodProxyAndGC]). 
	suite run. 
 ! !

!CompiledMethodSwapper classSide methodsFor: 'stuff to exclude' stamp: 'MarianoMartinezPeck 5/13/2011 13:58'!
packagesNotToSwap.
	"Notice that this is the strict way. If there are errors difficult to debug I would uncomment the commented ones and try."
	^ (# ("'System-Object Storage'" 'Fuel' 'Kernel' "'Graphics'" 'Exceptions' 'Proxies' "'Settings' ""'System'" 'UsedObject' "'Traits'" "'Balloon'" 'Tools-Base' "'Morphic'" "'ToolBuilder'" 'Compiler' ) collect: [:aPackageName | PackageInfo named: aPackageName]),
	(PackageInfo allPackages select: [:each | each packageName includesSubString: 'Collections-' ])
	
! !

!CompiledMethodSwapper classSide methodsFor: 'logging' stamp: 'MarianoMartinezPeck 12/2/2010 16:04'!
printSpaceTally
	self log: ''.
	self log: ''.
	SpaceTally new printSpaceAnalysis: 1 on: self logger logFile maxResults: 10. 
	self log: ''.
	self log: ''.! !

!CompiledMethodSwapper classSide methodsFor: 'accessing' stamp: 'MarianoMartinezPeck 11/21/2010 16:24'!
proxyClass
	^ ProxyClass! !

!CompiledMethodSwapper classSide methodsFor: 'accessing' stamp: 'MarianoMartinezPeck 11/21/2010 16:25'!
proxyClass: aProxyClass
	ProxyClass := aProxyClass.! !

!CompiledMethodSwapper classSide methodsFor: 'stuff to exclude' stamp: 'MarianoMartinezPeck 5/3/2011 18:06'!
selectorsNotToSwap
	^ #(#class #DoIt), (Smalltalk specialObjectsArray select: [:each | each isSymbol])
	
! !

Object subclass: #GraphTraverser
	instanceVariableNames: 'serializedObjects stack'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Marea-Swapping'!

GraphTraverser subclass: #GraphProxiesCreator
	instanceVariableNames: 'proxiesDict'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Marea-Swapping'!

!GraphProxiesCreator methodsFor: 'as yet unclassified' stamp: 'MarianoMartinezPeck 9/13/2011 10:54'!
createProxyFor: anObject
	^ MareaProxyMock new.! !

!GraphProxiesCreator methodsFor: 'as yet unclassified' stamp: 'MarianoMartinezPeck 9/13/2011 11:17'!
doInterception: anObject
	(anObject class = SmallInteger) 
		ifFalse: [ proxiesDict at: anObject put: (self createProxyFor: anObject)]! !

!GraphProxiesCreator methodsFor: 'as yet unclassified' stamp: 'MarianoMartinezPeck 9/13/2011 10:53'!
initialize 
	super initialize.
	proxiesDict := IdentityDictionary new.! !

!GraphProxiesCreator methodsFor: 'as yet unclassified' stamp: 'MarianoMartinezPeck 9/13/2011 11:00'!
proxiesDict
	^ proxiesDict ! !

!GraphProxiesCreator methodsFor: 'as yet unclassified' stamp: 'MarianoMartinezPeck 9/13/2011 11:06'!
traverse: anObject
	super traverse: anObject.
	^ proxiesDict! !

!GraphTraverser classSide methodsFor: 'as yet unclassified' stamp: 'MarianoMartinezPeck 9/13/2011 10:43'!
traverse: anObject
	^ self new traverse: anObject! !

!GraphTraverser methodsFor: 'as yet unclassified' stamp: 'MarianoMartinezPeck 9/13/2011 10:51'!
doInterception: anObject
	"By default do not do anything"! !

!GraphTraverser methodsFor: 'as yet unclassified' stamp: 'MarianoMartinezPeck 9/13/2011 10:43'!
followReferences: anObject

	(1 to: anObject class instSize) do: [:index | stack push: (anObject instVarAt: index)].

	1 to: anObject basicSize do: [ :index |
		stack push: (anObject basicAt: index) ]

! !

!GraphTraverser methodsFor: 'as yet unclassified' stamp: 'MarianoMartinezPeck 9/13/2011 10:39'!
initialize 
	serializedObjects := IdentitySet new.
	stack := FLSimpleStack new.! !

!GraphTraverser methodsFor: 'as yet unclassified' stamp: 'MarianoMartinezPeck 9/13/2011 10:50'!
trace: anObject

	serializedObjects addIfNotPresent: anObject ifPresentDo: [ ^self ].
	self doInterception: anObject.
	self followReferences: anObject! !

!GraphTraverser methodsFor: 'as yet unclassified' stamp: 'MarianoMartinezPeck 9/13/2011 10:40'!
traverse: anObject 

	stack push: anObject.

	[ stack isEmpty ] 
		whileFalse: [ self trace: stack pop ].
		
	^ serializedObjects
! !

Object subclass: #MareaProxyMock
	instanceVariableNames: 'position graphID'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Marea-Swapping'!

!MareaProxyMock methodsFor: 'accessing' stamp: 'MarianoMartinezPeck 9/14/2011 19:53'!
mareaGraphID
	^ graphID! !

!MareaProxyMock methodsFor: 'accessing' stamp: 'MarianoMartinezPeck 9/14/2011 20:01'!
mareaPosition
	^ position ! !

!MareaProxyMock methodsFor: 'as yet unclassified' stamp: 'MarianoMartinezPeck 9/9/2011 11:20'!
mareaSetGraphID: aNumber
	graphID := aNumber
	! !

!MareaProxyMock methodsFor: 'as yet unclassified' stamp: 'MarianoMartinezPeck 9/9/2011 11:19'!
mareaSetPosition: aPosition
	position := aPosition.
	! !

Object subclass: #ObjectGraphSwapper
	instanceVariableNames: ''
	classVariableNames: 'GraphTable IDGenerator Logger'
	poolDictionaries: ''
	category: 'Marea-Swapping'!

!ObjectGraphSwapper classSide methodsFor: 'as yet unclassified' stamp: 'MarianoMartinezPeck 9/9/2011 11:59'!
globalGraphTable
	GraphTable isNil ifTrue: [GraphTable := IdentityDictionary new].
	^ GraphTable! !

!ObjectGraphSwapper classSide methodsFor: 'as yet unclassified' stamp: 'MarianoMartinezPeck 9/9/2011 11:27'!
nextGraphID
	IDGenerator isNil ifTrue: [IDGenerator := 0].
	IDGenerator := IDGenerator + 1.
	^ IDGenerator! !

!ObjectGraphSwapper classSide methodsFor: 'as yet unclassified' stamp: 'MarianoMartinezPeck 9/9/2011 14:44'!
resetGlobalTable
	GraphTable := nil.! !

!ObjectGraphSwapper classSide methodsFor: 'as yet unclassified' stamp: 'MarianoMartinezPeck 9/9/2011 14:45'!
resetGraphID
	IDGenerator := nil.! !

!ObjectGraphSwapper methodsFor: 'as yet unclassified' stamp: 'MarianoMartinezPeck 9/9/2011 11:59'!
globalGraphTable
	^ self class globalGraphTable! !

!ObjectGraphSwapper methodsFor: 'as yet unclassified' stamp: 'MarianoMartinezPeck 9/9/2011 11:28'!
nextAvailableID
	^ self class nextGraphID ! !

!ObjectGraphSwapper methodsFor: 'as yet unclassified' stamp: 'MarianoMartinezPeck 9/13/2011 11:15'!
originalObjectsAbleToBecome: originalObjects! !

!ObjectGraphSwapper methodsFor: 'as yet unclassified' stamp: 'MarianoMartinezPeck 9/14/2011 20:04'!
swapInGraphFor: aProxy
	"So....a proxy has received a message and hence it needs to be uninstalled: it has to read the file from disc, materialize back the original graph and to the become back between all the proxies of the graph and the just materialized objects "

	| materialization graphID materializedGraph dict proxiesOfGraph |
	
	graphID := aProxy mareaGraphID.
	
	"We get the ID and we search a file named as the ID. Then we materialize it."	
	materialization := FLMaterializer fromFileNamed: graphID asString.
	materializedGraph := materialization run root.
	 
	"Now we need to become all the proxies of the graph to the new materialized objects"
	dict := IdentityDictionary new.
	proxiesOfGraph := self globalGraphTable at: graphID.
	proxiesOfGraph do: [:each | 
		"Since this collection is weak, if it is nil it means it was a proxy for an object (inner or shared) which was finally garbage collected"
		each ifNotNil: 
			[dict at: each put: (materialization materializedInstances at: each mareaPosition)]
		].
	dict keys elementsForwardIdentityTo: dict values.
	
	"Finally we clean the global table"
	self globalGraphTable removeKey: graphID
	! !

!ObjectGraphSwapper methodsFor: 'as yet unclassified' stamp: 'MarianoMartinezPeck 9/14/2011 19:49'!
swapOutGraph: anObject
	| dict serializer serializedInstances graphID serialization |

	"We traverse the graph and, for each object, we create a proxy for it. They are added to an IdentityDictionary. Each key of the dict is an original object of the graph. The value associated with each key, is the proxy created for that object"
	dict := GraphProxiesCreator traverse: anObject.
	
	"We serialize the graph"
	graphID := self nextAvailableID.
	serializer := FLSerializer newDefault.
	serialization := serializer 
		serialize: anObject 
		on: (StandardFileStream fileNamed: graphID asString) binary.
	
	"Now that the graph was serialized, I need to set to each proxy, its associated relative offset in the stream. This is necessary when we need to uninstall the proxy and the back the become. We need to know who to do the became with."
	serializedInstances := serialization objectsIndexes.
	dict keysAndValuesDo: [:originalObject :proxy |
		proxy mareaSetPosition: (serializedInstances at: originalObject ifAbsent: []) .	
		proxy mareaSetGraphID: graphID.	
	].

	"Finally, do a bulk become betwee the original objects and the proxies"
	dict keys elementsForwardIdentityTo: dict values.
	
	"There is a global table that keeps an entry per swapped graph. It is important to know which proxies there are for a certain graph, because when a proxy receives a message and it needs to load back the graph, it needs to known all the rest of the proxies that need the become as well. Hence, we keep a weak list of the proxies of the graph "
	self globalGraphTable at: graphID put: (WeakArray withAll: dict values).

	dict := nil.
	serializer := nil.
	serializedInstances := nil.
	"At this moment, having done the bulk become and set the dict to nil, there are several objects that can be garbage collected: ALL original objects of the graph, no matter whether they are inner or shared. And ALL proxies that were associated with inner objects. The only proxies that are reachable, hence not garbage collected, are the proxy for the root of the graph, and the proxies for facade objects. Of course, notice that the whole object memory graph can change, and for example, it may happen that nobody else point to the root (now proxy) of the graph or to facade objects. If that happens, then those proxies are garbage collected as well "
	
	! !

FLSerializer subclass: #FLSerializerForSwapping
	instanceVariableNames: 'weakSerializedObjects'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Marea-Serializers'!

!FLSerializerForSwapping methodsFor: 'as yet unclassified' stamp: 'MarianoMartinezPeck 8/2/2011 14:40'!
initialize
	super initialize.
	weakSerializedObjects := WeakOrderedCollection new.! !

!FLSerializerForSwapping methodsFor: 'as yet unclassified' stamp: 'MarianoMartinezPeck 8/3/2011 18:44'!
serializeCluster: aCluster objects: aCollection

	aCollection do: [ :instance | 
		instanceIndexes at: instance put: instanceIndexes size + 1.
		(instance == root) ifTrue: [ weakSerializedObjects addLast: nil ]
		ifFalse: [weakSerializedObjects addLast: instance].
		 ].
	self encodeIdOf: aCluster.
	aCluster serializeAll: aCollection with: self.
! !

!FLSerializerForSwapping methodsFor: 'as yet unclassified' stamp: 'MarianoMartinezPeck 8/2/2011 11:18'!
weakSerializedObjects 
	^ weakSerializedObjects ! !
